@@ -4,13 +4,12 @@ use ObjStore ':ALL';
 set_transaction_priority(0);
 set_max_retries(0);
 use lib "./t";
+use test;
 
-open(STDOUT, ">>/dev/null") or die "open: $@";
-open(STDERR, ">>/dev/null") or die "open: $@";
-
-my $osdir = ObjStore->schema_dir;
-my $db = ObjStore::open($osdir . "/perltest.db", 0, 0666);
+#open(STDOUT, ">>/dev/null") or die "open: $@";
+#open(STDERR, ">>/dev/null") or die "open: $@";
 
+&open_db;
 try_update {
     my $left = $db->root('tripwire');
     ++ $left->{left};
